#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: ipmi_boot
short_description: Management of order of boot devices
description:
  - Use this module to manage order of boot devices
options:
  name:
    description:
      - Hostname or ip address of the BMC.
    required: true
    type: str
  port:
    description:
      - Remote RMCP port.
    default: 623
    type: int
  user:
    description:
      - Username to use to connect to the BMC.
    required: true
    type: str
  password:
    description:
      - Password to connect to the BMC.
    required: true
    type: str
  key:
    description:
      - Encryption key to connect to the BMC in hex format.
    required: false
    type: str
    version_added: 4.1.0
  bootdev:
    description:
      - Set boot device to use on next reboot
      - "The choices for the device are:
          - network -- Request network boot
          - floppy -- Boot from floppy
          - hd -- Boot from hard drive
          - safe -- Boot from hard drive, requesting 'safe mode'
          - optical -- boot from CD/DVD/BD drive
          - setup -- Boot into setup utility
          - default -- remove any IPMI directed boot device request"
    required: true
    choices:
      - network
      - floppy
      - hd
      - safe
      - optical
      - setup
      - default
    type: str
  state:
    description:
      - Whether to ensure that boot devices is desired.
      - "The choices for the state are:
            - present -- Request system turn on
            - absent -- Request system turn on"
    default: present
    choices: [ present, absent ]
    type: str
  persistent:
    description:
      - If set, ask that system firmware uses this device beyond next boot.
        Be aware many systems do not honor this.
    type: bool
    default: 'no'
  uefiboot:
    description:
      - If set, request UEFI boot explicitly.
        Strictly speaking, the spec suggests that if not set, the system should BIOS boot and offers no "don't care" option.
        In practice, this flag not being set does not preclude UEFI boot on any system I've encountered.
    type: bool
    default: 'no'
requirements:
  - "python >= 2.6"
  - pyghmi
author: "Bulat Gaifullin (@bgaifullin) <gaifullinbf@gmail.com>"
'''

RETURN = '''
bootdev:
    description: The boot device name which will be used beyond next boot.
    returned: success
    type: str
    sample: default
persistent:
    description: If True, system firmware will use this device beyond next boot.
    returned: success
    type: bool
    sample: false
uefimode:
    description: If True, system firmware will use UEFI boot explicitly beyond next boot.
    returned: success
    type: bool
    sample: false
'''

EXAMPLES = '''
- name: Ensure bootdevice is HD
  community.general.ipmi_boot:
    name: test.testdomain.com
    user: admin
    password: password
    bootdev: hd

- name: Ensure bootdevice is not Network
  community.general.ipmi_boot:
    name: test.testdomain.com
    user: admin
    password: password
    key: 1234567890AABBCCDEFF000000EEEE12
    bootdev: network
    state: absent
'''

import traceback
import binascii

PYGHMI_IMP_ERR = None
try:
    from pyghmi.ipmi import command
except ImportError:
    PYGHMI_IMP_ERR = traceback.format_exc()
    command = None

from ansible.module_utils.basic import AnsibleModule, missing_required_lib


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(required=True),
            port=dict(default=623, type='int'),
            user=dict(required=True, no_log=True),
            password=dict(required=True, no_log=True),
            key=dict(type='str', no_log=True),
            state=dict(default='present', choices=['present', 'absent']),
            bootdev=dict(required=True, choices=['network', 'hd', 'floppy', 'safe', 'optical', 'setup', 'default']),
            persistent=dict(default=False, type='bool'),
            uefiboot=dict(default=False, type='bool')
        ),
        supports_check_mode=True,
    )

    if command is None:
        module.fail_json(msg=missing_required_lib('pyghmi'), exception=PYGHMI_IMP_ERR)

    name = module.params['name']
    port = module.params['port']
    user = module.params['user']
    password = module.params['password']
    state = module.params['state']
    bootdev = module.params['bootdev']
    persistent = module.params['persistent']
    uefiboot = module.params['uefiboot']
    request = dict()

    if state == 'absent' and bootdev == 'default':
        module.fail_json(msg="The bootdev 'default' cannot be used with state 'absent'.")

    try:
        if module.params['key']:
            key = binascii.unhexlify(module.params['key'])
        else:
            key = None
    except Exception as e:
        module.fail_json(msg="Unable to convert 'key' from hex string.")

    # --- run command ---
    try:
        ipmi_cmd = command.Command(
            bmc=name, userid=user, password=password, port=port, kg=key
        )
        module.debug('ipmi instantiated - name: "%s"' % name)
        current = ipmi_cmd.get_bootdev()
        # uefimode may not supported by BMC, so use desired value as default
        current.setdefault('uefimode', uefiboot)
        if state == 'present' and current != dict(bootdev=bootdev, persistent=persistent, uefimode=uefiboot):
            request = dict(bootdev=bootdev, uefiboot=uefiboot, persist=persistent)
        elif state == 'absent' and current['bootdev'] == bootdev:
            request = dict(bootdev='default')
        else:
            module.exit_json(changed=False, **current)

        if module.check_mode:
            response = dict(bootdev=request['bootdev'])
        else:
            response = ipmi_cmd.set_bootdev(**request)

        if 'error' in response:
            module.fail_json(msg=response['error'])

        if 'persist' in request:
            response['persistent'] = request['persist']
        if 'uefiboot' in request:
            response['uefimode'] = request['uefiboot']

        module.exit_json(changed=True, **response)
    except Exception as e:
        module.fail_json(msg=str(e))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright: (c) 2021, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

from contextlib import contextmanager

from ansible_collections.community.general.tests.unit.compat import unittest
from ansible_collections.community.general.tests.unit.compat.mock import call, patch
from ansible_collections.community.general.tests.unit.plugins.modules.utils import AnsibleExitJson, AnsibleFailJson, \
    ModuleTestCase, set_module_args

from ansible_collections.community.general.plugins.modules.identity.keycloak import keycloak_clientscope

from itertools import count

from ansible.module_utils.six import StringIO


@contextmanager
def patch_keycloak_api(get_clientscope_by_name=None, get_clientscope_by_clientscopeid=None, create_clientscope=None,
                       update_clientscope=None, get_clientscope_protocolmapper_by_name=None,
                       update_clientscope_protocolmappers=None, create_clientscope_protocolmapper=None,
                       delete_clientscope=None):
    """Mock context manager for patching the methods in PwPolicyIPAClient that contact the IPA server

    Patches the `login` and `_post_json` methods

    Keyword arguments are passed to the mock object that patches `_post_json`

    No arguments are passed to the mock object that patches `login` because no tests require it

    Example::

        with patch_ipa(return_value={}) as (mock_login, mock_post):
            ...
    """

    """
    get_clientscope_by_clientscopeid
    delete_clientscope
    """

    obj = keycloak_clientscope.KeycloakAPI
    with patch.object(obj, 'get_clientscope_by_name', side_effect=get_clientscope_by_name) \
            as mock_get_clientscope_by_name:
        with patch.object(obj, 'get_clientscope_by_clientscopeid', side_effect=get_clientscope_by_clientscopeid) \
                as mock_get_clientscope_by_clientscopeid:
            with patch.object(obj, 'create_clientscope', side_effect=create_clientscope) \
                    as mock_create_clientscope:
                with patch.object(obj, 'update_clientscope', return_value=update_clientscope) \
                        as mock_update_clientscope:
                    with patch.object(obj, 'get_clientscope_protocolmapper_by_name',
                                      side_effect=get_clientscope_protocolmapper_by_name) \
                            as mock_get_clientscope_protocolmapper_by_name:
                        with patch.object(obj, 'update_clientscope_protocolmappers',
                                          side_effect=update_clientscope_protocolmappers) \
                                as mock_update_clientscope_protocolmappers:
                            with patch.object(obj, 'create_clientscope_protocolmapper',
                                              side_effect=create_clientscope_protocolmapper) \
                                    as mock_create_clientscope_protocolmapper:
                                with patch.object(obj, 'delete_clientscope', side_effect=delete_clientscope) \
                                        as mock_delete_clientscope:
                                    yield mock_get_clientscope_by_name, mock_get_clientscope_by_clientscopeid, mock_create_clientscope, \
                                        mock_update_clientscope, mock_get_clientscope_protocolmapper_by_name, mock_update_clientscope_protocolmappers, \
                                        mock_create_clientscope_protocolmapper, mock_delete_clientscope


def get_response(object_with_future_response, method, get_id_call_count):
    if callable(object_with_future_response):
        return object_with_future_response()
    if isinstance(object_with_future_response, dict):
        return get_response(
            object_with_future_response[method], method, get_id_call_count)
    if isinstance(object_with_future_response, list):
        call_number = next(get_id_call_count)
        return get_response(
            object_with_future_response[call_number], method, get_id_call_count)
    return object_with_future_response


def build_mocked_request(get_id_user_count, response_dict):
    def _mocked_requests(*args, **kwargs):
        url = args[0]
        method = kwargs['method']
        future_response = response_dict.get(url, None)
        return get_response(future_response, method, get_id_user_count)

    return _mocked_requests


def create_wrapper(text_as_string):
    """Allow to mock many times a call to one address.
    Without this function, the StringIO is empty for the second call.
    """

    def _create_wrapper():
        return StringIO(text_as_string)

    return _create_wrapper


def mock_good_connection():
    token_response = {
        'http://keycloak.url/auth/realms/master/protocol/openid-connect/token': create_wrapper(
            '{"access_token": "alongtoken"}'), }
    return patch(
        'ansible_collections.community.general.plugins.module_utils.identity.keycloak.keycloak.open_url',
        side_effect=build_mocked_request(count(), token_response),
        autospec=True
    )


class TestKeycloakAuthentication(ModuleTestCase):
    def setUp(self):
        super(TestKeycloakAuthentication, self).setUp()
        self.module = keycloak_clientscope

    def test_create_clientscope(self):
        """Add a new authentication flow from copy of an other flow"""

        module_args = {
            'auth_keycloak_url': 'http://keycloak.url/auth',
            'auth_username': 'admin',
            'auth_password': 'admin',
            'auth_realm': 'master',
            'realm': 'realm-name',
            'state': 'present',
            'name': 'my-new-kc-clientscope'
        }
        return_value_get_clientscope_by_name = [
            None,
            {
                "attributes": {},
                "id": "73fec1d2-f032-410c-8177-583104d01305",
                "name": "my-new-kc-clientscope"
            }]

        changed = True

        set_module_args(module_args)

        # Run the module

        with mock_good_connection():
            with patch_keycloak_api(get_clientscope_by_name=return_value_get_clientscope_by_name) \
                    as (mock_get_clientscope_by_name, mock_get_clientscope_by_clientscopeid, mock_create_clientscope,
                        mock_update_clientscope, mock_get_clientscope_protocolmapper_by_name,
                        mock_update_clientscope_protocolmappers,
                        mock_create_clientscope_protocolmapper, mock_delete_clientscope):
                with self.assertRaises(AnsibleExitJson) as exec_info:
                    self.module.main()

        # Verify number of call on each mock
        self.assertEqual(mock_get_clientscope_by_name.call_count, 2)
        self.assertEqual(mock_create_clientscope.call_count, 1)
        self.assertEqual(mock_get_clientscope_by_clientscopeid.call_count, 0)
        self.assertEqual(mock_update_clientscope.call_count, 0)
        self.assertEqual(mock_get_clientscope_protocolmapper_by_name.call_count, 0)
        self.assertEqual(mock_update_clientscope_protocolmappers.call_count, 0)
        self.assertEqual(mock_create_clientscope_protocolmapper.call_count, 0)
        self.assertEqual(mock_delete_clientscope.call_count, 0)

        # Verify that the module's changed status matches what is expected
        self.assertIs(exec_info.exception.args[0]['changed'], changed)

    def test_create_clientscope_idempotency(self):
        """Add a new authentication flow from copy of an other flow"""

        module_args = {
            'auth_keycloak_url': 'http://keycloak.url/auth',
            'auth_username': 'admin',
            'auth_password': 'admin',
            'auth_realm': 'master',
            'realm': 'realm-name',
            'state': 'present',
            'name': 'my-new-kc-clientscope'
        }
        return_value_get_clientscope_by_name = [{
            "attributes": {},
            "id": "73fec1d2-f032-410c-8177-583104d01305",
            "name": "my-new-kc-clientscope"
        }]

        changed = False

        set_module_args(module_args)

        # Run the module

        with mock_good_connection():
            with patch_keycloak_api(get_clientscope_by_name=return_value_get_clientscope_by_name) \
                    as (mock_get_clientscope_by_name, mock_get_clientscope_by_clientscopeid, mock_create_clientscope,
                        mock_update_clientscope, mock_get_clientscope_protocolmapper_by_name,
                        mock_update_clientscope_protocolmappers,
                        mock_create_clientscope_protocolmapper, mock_delete_clientscope):
                with self.assertRaises(AnsibleExitJson) as exec_info:
                    self.module.main()

        # Verify number of call on each mock
        self.assertEqual(mock_get_clientscope_by_name.call_count, 1)
        self.assertEqual(mock_create_clientscope.call_count, 0)
        self.assertEqual(mock_get_clientscope_by_clientscopeid.call_count, 0)
        self.assertEqual(mock_update_clientscope.call_count, 0)
        self.assertEqual(mock_get_clientscope_protocolmapper_by_name.call_count, 0)
        self.assertEqual(mock_update_clientscope_protocolmappers.call_count, 0)
        self.assertEqual(mock_create_clientscope_protocolmapper.call_count, 0)
        self.assertEqual(mock_delete_clientscope.call_count, 0)

        # Verify that the module's changed status matches what is expected
        self.assertIs(exec_info.exception.args[0]['changed'], changed)

    def test_delete_clientscope(self):
        """Add a new authentication flow from copy of an other flow"""

        module_args = {
            'auth_keycloak_url': 'http://keycloak.url/auth',
            'auth_username': 'admin',
            'auth_password': 'admin',
            'auth_realm': 'master',
            'realm': 'realm-name',
            'state': 'absent',
            'name': 'my-new-kc-clientscope'
        }
        return_value_get_clientscope_by_name = [{
            "attributes": {},
            "id": "73fec1d2-f032-410c-8177-583104d01305",
            "name": "my-new-kc-clientscope"
        }]

        changed = True

        set_module_args(module_args)

        # Run the module

        with mock_good_connection():
            with patch_keycloak_api(get_clientscope_by_name=return_value_get_clientscope_by_name) \
                    as (mock_get_clientscope_by_name, mock_get_clientscope_by_clientscopeid, mock_create_clientscope,
                        mock_update_clientscope, mock_get_clientscope_protocolmapper_by_name,
                        mock_update_clientscope_protocolmappers,
                        mock_create_clientscope_protocolmapper, mock_delete_clientscope):
                with self.assertRaises(AnsibleExitJson) as exec_info:
                    self.module.main()

        # Verify number of call on each mock
        self.assertEqual(mock_get_clientscope_by_name.call_count, 1)
        self.assertEqual(mock_create_clientscope.call_count, 0)
        self.assertEqual(mock_get_clientscope_by_clientscopeid.call_count, 0)
        self.assertEqual(mock_update_clientscope.call_count, 0)
        self.assertEqual(mock_get_clientscope_protocolmapper_by_name.call_count, 0)
        self.assertEqual(mock_update_clientscope_protocolmappers.call_count, 0)
        self.assertEqual(mock_create_clientscope_protocolmapper.call_count, 0)
        self.assertEqual(mock_delete_clientscope.call_count, 1)

        # Verify that the module's changed status matches what is expected
        self.assertIs(exec_info.exception.args[0]['changed'], changed)

    def test_delete_clientscope_idempotency(self):
        """Add a new authentication flow from copy of an other flow"""

        module_args = {
            'auth_keycloak_url': 'http://keycloak.url/auth',
            'auth_username': 'admin',
            'auth_password': 'admin',
            'auth_realm': 'master',
            'realm': 'realm-name',
            'state': 'absent',
            'name': 'my-new-kc-clientscope'
        }
        return_value_get_clientscope_by_name = [None]

        changed = False

        set_module_args(module_args)

        # Run the module

        with mock_good_connection():
            with patch_keycloak_api(get_clientscope_by_name=return_value_get_clientscope_by_name) \
                    as (mock_get_clientscope_by_name, mock_get_clientscope_by_clientscopeid, mock_create_clientscope,
                        mock_update_clientscope, mock_get_clientscope_protocolmapper_by_name,
                        mock_update_clientscope_protocolmappers,
                        mock_create_clientscope_protocolmapper, mock_delete_clientscope):
                with self.assertRaises(AnsibleExitJson) as exec_info:
                    self.module.main()

        # Verify number of call on each mock
        self.assertEqual(mock_get_clientscope_by_name.call_count, 1)
        self.assertEqual(mock_create_clientscope.call_count, 0)
        self.assertEqual(mock_get_clientscope_by_clientscopeid.call_count, 0)
        self.assertEqual(mock_update_clientscope.call_count, 0)
        self.assertEqual(mock_get_clientscope_protocolmapper_by_name.call_count, 0)
        self.assertEqual(mock_update_clientscope_protocolmappers.call_count, 0)
        self.assertEqual(mock_create_clientscope_protocolmapper.call_count, 0)
        self.assertEqual(mock_delete_clientscope.call_count, 0)

        # Verify that the module's changed status matches what is expected
        self.assertIs(exec_info.exception.args[0]['changed'], changed)

    def test_create_clientscope_with_protocolmappers(self):
        """Add a new authentication flow from copy of an other flow"""

        module_args = {
            'auth_keycloak_url': 'http://keycloak.url/auth',
            'auth_username': 'admin',
            'auth_password': 'admin',
            'auth_realm': 'master',
            'realm': 'realm-name',
            'state': 'present',
            'name': 'my-new-kc-clientscope',
            'protocolMappers': [
                {
                    'protocol': 'openid-connect',
                    'config': {
                        'full.path': 'true',
                        'id.token.claim': 'true',
                        'access.token.claim': 'true',
                        'userinfo.token.claim': 'true',
                        'claim.name': 'protocol1',
                    },
                    'name': 'protocol1',
                    'protocolMapper': 'oidc-group-membership-mapper',
                },
                {
                    'protocol': 'openid-connect',
                    'config': {
                        'full.path': 'false',
                        'id.token.claim': 'false',
                        'access.token.claim': 'false',
                        'userinfo.token.claim': 'false',
                        'claim.name': 'protocol2',
                    },
                    'name': 'protocol2',
                    'protocolMapper': 'oidc-group-membership-mapper',
                },
                {
                    'protocol': 'openid-connect',
                    'config': {
                        'full.path': 'true',
                        'id.token.claim': 'false',
                        'access.token.claim': 'true',
                        'userinfo.token.claim': 'false',
                        'claim.name': 'protocol3',
                    },
                    'name': 'protocol3',
                    'protocolMapper': 'oidc-group-membership-mapper',
                },
            ]
        }
        return_value_get_clientscope_by_name = [
            None,
            {
                "attributes": {},
                "id": "890ec72e-fe1d-4308-9f27-485ef7eaa182",
                "name": "my-new-kc-clientscope",
                "protocolMappers": [
                    {
                        "config": {
                            "access.token.claim": "false",
                            "claim.name": "protocol2",
                            "full.path": "false",
                            "id.token.claim": "false",
                            "userinfo.token.claim": "false"
                        },
                        "consentRequired": "false",
                        "id": "a7f19adb-cc58-41b1-94ce-782dc255139b",
                        "name": "protocol2",
                        "protocol": "openid-connect",
                        "protocolMapper": "oidc-group-membership-mapper"
                    },
                    {
                        "config": {
                            "access.token.claim": "true",
                            "claim.name": "protocol3",
                            "full.path": "true",
                            "id.token.claim": "false",
                            "userinfo.token.claim": "false"
                        },
                        "consentRequired": "false",
                        "id": "2103a559-185a-40f4-84ae-9ab311d5b812",
                        "name": "protocol3",
                        "protocol": "openid-connect",
                        "protocolMapper": "oidc-group-membership-mapper"
                    },
                    {
                        "config": {
                            "access.token.claim": "true",
                            "claim.name": "protocol1",
                            "full.path": "true",
                            "id.token.claim": "true",
                            "userinfo.token.claim": "true"
                        },
                        "consentRequired": "false",
                        "id": "bbf6390f-e95f-4c20-882b-9dad328363b9",
                        "name": "protocol1",
                        "protocol": "openid-connect",
                        "protocolMapper": "oidc-group-membership-mapper"
                    }]
            }]

        changed = True

        set_module_args(module_args)

        # Run the module

        with mock_good_connection():
            with patch_keycloak_api(get_clientscope_by_name=return_value_get_clientscope_by_name) \
                    as (mock_get_clientscope_by_name, mock_get_clientscope_by_clientscopeid, mock_create_clientscope,
                        mock_update_clientscope, mock_get_clientscope_protocolmapper_by_name,
                        mock_update_clientscope_protocolmappers,
                        mock_create_clientscope_protocolmapper, mock_delete_clientscope):
                with self.assertRaises(AnsibleExitJson) as exec_info:
                    self.module.main()

        # Verify number of call on each mock
        self.assertEqual(mock_get_clientscope_by_name.call_count, 2)
        self.assertEqual(mock_create_clientscope.call_count, 1)
        self.assertEqual(mock_get_clientscope_by_clientscopeid.call_count, 0)
        self.assertEqual(mock_update_clientscope.call_count, 0)
        self.assertEqual(mock_get_clientscope_protocolmapper_by_name.call_count, 0)
        self.assertEqual(mock_update_clientscope_protocolmappers.call_count, 0)
        self.assertEqual(mock_create_clientscope_protocolmapper.call_count, 0)
        self.assertEqual(mock_delete_clientscope.call_count, 0)

        # Verify that the module's changed status matches what is expected
        self.assertIs(exec_info.exception.args[0]['changed'], changed)

    def test_update_clientscope_with_protocolmappers(self):
        """Add a new authentication flow from copy of an other flow"""

        module_args = {
            'auth_keycloak_url': 'http://keycloak.url/auth',
            'auth_username': 'admin',
            'auth_password': 'admin',
            'auth_realm': 'master',
            'realm': 'realm-name',
            'state': 'present',
            'name': 'my-new-kc-clientscope',
            'protocolMappers': [
                {
                    'protocol': 'openid-connect',
                    'config': {
                        'full.path': 'false',
                        'id.token.claim': 'false',
                        'access.token.claim': 'false',
                        'userinfo.token.claim': 'false',
                        'claim.name': 'protocol1_updated',
                    },
                    'name': 'protocol1',
                    'protocolMapper': 'oidc-group-membership-mapper',
                },
                {
                    'protocol': 'openid-connect',
                    'config': {
                        'full.path': 'true',
                        'id.token.claim': 'false',
                        'access.token.claim': 'false',
                        'userinfo.token.claim': 'false',
                        'claim.name': 'protocol2_updated',
                    },
                    'name': 'protocol2',
                    'protocolMapper': 'oidc-group-membership-mapper',
                },
                {
                    'protocol': 'openid-connect',
                    'config': {
                        'full.path': 'true',
                        'id.token.claim': 'true',
                        'access.token.claim': 'true',
                        'userinfo.token.claim': 'true',
                        'claim.name': 'protocol3_updated',
                    },
                    'name': 'protocol3',
                    'protocolMapper': 'oidc-group-membership-mapper',
                },
            ]
        }
        return_value_get_clientscope_by_name = [{
            "attributes": {},
            "id": "890ec72e-fe1d-4308-9f27-485ef7eaa182",
            "name": "my-new-kc-clientscope",
            "protocolMappers": [
                {
                    "config": {
                        "access.token.claim": "true",
                        "claim.name": "groups",
                        "full.path": "true",
                        "id.token.claim": "true",
                        "userinfo.token.claim": "true"
                    },
                    "consentRequired": "false",
                    "id": "e077007a-367a-444f-91ef-70277a1d868d",
                    "name": "groups",
                    "protocol": "saml",
                    "protocolMapper": "oidc-group-membership-mapper"
                },
                {
                    "config": {
                        "access.token.claim": "true",
                        "claim.name": "groups",
                        "full.path": "true",
                        "id.token.claim": "true",
                        "userinfo.token.claim": "true"
                    },
                    "consentRequired": "false",
                    "id": "06c518aa-c627-43cc-9a82-d8467b508d34",
                    "name": "groups",
                    "protocol": "openid-connect",
                    "protocolMapper": "oidc-group-membership-mapper"
                },
                {
                    "config": {
                        "access.token.claim": "true",
                        "claim.name": "groups",
                        "full.path": "true",
                        "id.token.claim": "true",
                        "userinfo.token.claim": "true"
                    },
                    "consentRequired": "false",
                    "id": "1d03c557-d97e-40f4-ac35-6cecd74ea70d",
                    "name": "groups",
                    "protocol": "wsfed",
                    "protocolMapper": "oidc-group-membership-mapper"
                }
            ]
        }]
        return_value_get_clientscope_by_clientscopeid = [{
            "attributes": {},
            "id": "2286032f-451e-44d5-8be6-e45aac7983a1",
            "name": "my-new-kc-clientscope",
            "protocolMappers": [
                {
                    "config": {
                        "access.token.claim": "true",
                        "claim.name": "protocol1_updated",
                        "full.path": "true",
                        "id.token.claim": "false",
                        "userinfo.token.claim": "false"
                    },
                    "consentRequired": "false",
                    "id": "a7f19adb-cc58-41b1-94ce-782dc255139b",
                    "name": "protocol2",
                    "protocol": "openid-connect",
                    "protocolMapper": "oidc-group-membership-mapper"
                },
                {
                    "config": {
                        "access.token.claim": "true",
                        "claim.name": "protocol1_updated",
                        "full.path": "true",
                        "id.token.claim": "false",
                        "userinfo.token.claim": "false"
                    },
                    "consentRequired": "false",
                    "id": "2103a559-185a-40f4-84ae-9ab311d5b812",
                    "name": "protocol3",
                    "protocol": "openid-connect",
                    "protocolMapper": "oidc-group-membership-mapper"
                },
                {
                    "config": {
                        "access.token.claim": "false",
                        "claim.name": "protocol1_updated",
                        "full.path": "false",
                        "id.token.claim": "false",
                        "userinfo.token.claim": "false"
                    },
                    "consentRequired": "false",
                    "id": "bbf6390f-e95f-4c20-882b-9dad328363b9",
                    "name": "protocol1",
                    "protocol": "openid-connect",
                    "protocolMapper": "oidc-group-membership-mapper"
                }
            ]
        }]

        changed = True

        set_module_args(module_args)

        # Run the module

        with mock_good_connection():
            with patch_keycloak_api(get_clientscope_by_name=return_value_get_clientscope_by_name,
                                    get_clientscope_by_clientscopeid=return_value_get_clientscope_by_clientscopeid) \
                    as (mock_get_clientscope_by_name, mock_get_clientscope_by_clientscopeid, mock_create_clientscope,
                        mock_update_clientscope, mock_get_clientscope_protocolmapper_by_name,
                        mock_update_clientscope_protocolmappers,
                        mock_create_clientscope_protocolmapper, mock_delete_clientscope):
                with self.assertRaises(AnsibleExitJson) as exec_info:
                    self.module.main()

        # Verify number of call on each mock
        self.assertEqual(mock_get_clientscope_by_name.call_count, 1)
        self.assertEqual(mock_create_clientscope.call_count, 0)
        self.assertEqual(mock_get_clientscope_by_clientscopeid.call_count, 1)
        self.assertEqual(mock_update_clientscope.call_count, 1)
        self.assertEqual(mock_get_clientscope_protocolmapper_by_name.call_count, 3)
        self.assertEqual(mock_update_clientscope_protocolmappers.call_count, 3)
        self.assertEqual(mock_create_clientscope_protocolmapper.call_count, 0)
        self.assertEqual(mock_delete_clientscope.call_count, 0)

        # Verify that the module's changed status matches what is expected
        self.assertIs(exec_info.exception.args[0]['changed'], changed)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2016, Jiri Tyr <jiri.tyr@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: jenkins_plugin
author: Jiri Tyr (@jtyr)
short_description: Add or remove Jenkins plugin
description:
  - Ansible module which helps to manage Jenkins plugins.

options:
  group:
    type: str
    description:
      - Name of the Jenkins group on the OS.
    default: jenkins
  jenkins_home:
    type: path
    description:
      - Home directory of the Jenkins user.
    default: /var/lib/jenkins
  mode:
    type: raw
    description:
      - File mode applied on versioned plugins.
    default: '0644'
  name:
    type: str
    description:
      - Plugin name.
    required: yes
  owner:
    type: str
    description:
      - Name of the Jenkins user on the OS.
    default: jenkins
  state:
    type: str
    description:
      - Desired plugin state.
      - If the C(latest) is set, the check for new version will be performed
        every time. This is suitable to keep the plugin up-to-date.
    choices: [absent, present, pinned, unpinned, enabled, disabled, latest]
    default: present
  timeout:
    type: int
    description:
      - Server connection timeout in secs.
    default: 30
  updates_expiration:
    type: int
    description:
      - Number of seconds after which a new copy of the I(update-center.json)
        file is downloaded. This is used to avoid the need to download the
        plugin to calculate its checksum when C(latest) is specified.
      - Set it to C(0) if no cache file should be used. In that case, the
        plugin file will always be downloaded to calculate its checksum when
        C(latest) is specified.
    default: 86400
  updates_url:
    type: list
    elements: str
    description:
      - A list of base URL(s) to retrieve I(update-center.json), and direct plugin files from.
      - This can be a list since community.general 3.3.0.
    default: ['https://updates.jenkins.io', 'http://mirrors.jenkins.io']
  update_json_url_segment:
    type: list
    elements: str
    description:
      - A list of URL segment(s) to retrieve the update center json file from.
    default: ['update-center.json', 'updates/update-center.json']
    version_added: 3.3.0
  latest_plugins_url_segments:
    type: list
    elements: str
    description:
      - Path inside the I(updates_url) to get latest plugins from.
    default: ['latest']
    version_added: 3.3.0
  versioned_plugins_url_segments:
    type: list
    elements: str
    description:
      - Path inside the I(updates_url) to get specific version of plugins from.
    default: ['download/plugins', 'plugins']
    version_added: 3.3.0
  url:
    type: str
    description:
      - URL of the Jenkins server.
    default: http://localhost:8080
  version:
    type: str
    description:
      - Plugin version number.
      - If this option is specified, all plugin dependencies must be installed
        manually.
      - It might take longer to verify that the correct version is installed.
        This is especially true if a specific version number is specified.
      - Quote the version to prevent the value to be interpreted as float. For
        example if C(1.20) would be unquoted, it would become C(1.2).
  with_dependencies:
    description:
      - Defines whether to install plugin dependencies.
      - This option takes effect only if the I(version) is not defined.
    type: bool
    default: yes

notes:
  - Plugin installation should be run under root or the same user which owns
    the plugin files on the disk. Only if the plugin is not installed yet and
    no version is specified, the API installation is performed which requires
    only the Web UI credentials.
  - It's necessary to notify the handler or call the I(service) module to
    restart the Jenkins service after a new plugin was installed.
  - Pinning works only if the plugin is installed and Jenkins service was
    successfully restarted after the plugin installation.
  - It is not possible to run the module remotely by changing the I(url)
    parameter to point to the Jenkins server. The module must be used on the
    host where Jenkins runs as it needs direct access to the plugin files.
extends_documentation_fragment:
  - url
  - files
'''

EXAMPLES = '''
- name: Install plugin
  community.general.jenkins_plugin:
    name: build-pipeline-plugin

- name: Install plugin without its dependencies
  community.general.jenkins_plugin:
    name: build-pipeline-plugin
    with_dependencies: no

- name: Make sure the plugin is always up-to-date
  community.general.jenkins_plugin:
    name: token-macro
    state: latest

- name: Install specific version of the plugin
  community.general.jenkins_plugin:
    name: token-macro
    version: "1.15"

- name: Pin the plugin
  community.general.jenkins_plugin:
    name: token-macro
    state: pinned

- name: Unpin the plugin
  community.general.jenkins_plugin:
    name: token-macro
    state: unpinned

- name: Enable the plugin
  community.general.jenkins_plugin:
    name: token-macro
    state: enabled

- name: Disable the plugin
  community.general.jenkins_plugin:
    name: token-macro
    state: disabled

- name: Uninstall plugin
  community.general.jenkins_plugin:
    name: build-pipeline-plugin
    state: absent

#
# Example of how to authenticate
#
- name: Install plugin
  community.general.jenkins_plugin:
    name: build-pipeline-plugin
    url_username: admin
    url_password: p4ssw0rd
    url: http://localhost:8888

#
# Example of a Play which handles Jenkins restarts during the state changes
#
- name: Jenkins Master play
  hosts: jenkins-master
  vars:
    my_jenkins_plugins:
      token-macro:
        enabled: yes
      build-pipeline-plugin:
        version: "1.4.9"
        pinned: no
        enabled: yes
  tasks:
    - name: Install plugins without a specific version
      community.general.jenkins_plugin:
        name: "{{ item.key }}"
      register: my_jenkins_plugin_unversioned
      when: >
        'version' not in item.value
      with_dict: "{{ my_jenkins_plugins }}"

    - name: Install plugins with a specific version
      community.general.jenkins_plugin:
        name: "{{ item.key }}"
        version: "{{ item.value['version'] }}"
      register: my_jenkins_plugin_versioned
      when: >
        'version' in item.value
      with_dict: "{{ my_jenkins_plugins }}"

    - name: Initiate the fact
      ansible.builtin.set_fact:
        jenkins_restart_required: no

    - name: Check if restart is required by any of the versioned plugins
      ansible.builtin.set_fact:
        jenkins_restart_required: yes
      when: item.changed
      with_items: "{{ my_jenkins_plugin_versioned.results }}"

    - name: Check if restart is required by any of the unversioned plugins
      ansible.builtin.set_fact:
        jenkins_restart_required: yes
      when: item.changed
      with_items: "{{ my_jenkins_plugin_unversioned.results }}"

    - name: Restart Jenkins if required
      ansible.builtin.service:
        name: jenkins
        state: restarted
      when: jenkins_restart_required

    - name: Wait for Jenkins to start up
      ansible.builtin.uri:
        url: http://localhost:8080
        status_code: 200
        timeout: 5
      register: jenkins_service_status
      # Keep trying for 5 mins in 5 sec intervals
      retries: 60
      delay: 5
      until: >
         'status' in jenkins_service_status and
         jenkins_service_status['status'] == 200
      when: jenkins_restart_required

    - name: Reset the fact
      ansible.builtin.set_fact:
        jenkins_restart_required: no
      when: jenkins_restart_required

    - name: Plugin pinning
      community.general.jenkins_plugin:
        name: "{{ item.key }}"
        state: "{{ 'pinned' if item.value['pinned'] else 'unpinned'}}"
      when: >
        'pinned' in item.value
      with_dict: "{{ my_jenkins_plugins }}"

    - name: Plugin enabling
      community.general.jenkins_plugin:
        name: "{{ item.key }}"
        state: "{{ 'enabled' if item.value['enabled'] else 'disabled'}}"
      when: >
        'enabled' in item.value
      with_dict: "{{ my_jenkins_plugins }}"
'''

RETURN = '''
plugin:
    description: plugin name
    returned: success
    type: str
    sample: build-pipeline-plugin
state:
    description: state of the target, after execution
    returned: success
    type: str
    sample: "present"
'''

from ansible.module_utils.basic import AnsibleModule, to_bytes
from ansible.module_utils.six.moves import http_cookiejar as cookiejar
from ansible.module_utils.six.moves.urllib.parse import urlencode
from ansible.module_utils.urls import fetch_url, url_argument_spec
from ansible.module_utils.six import text_type, binary_type
from ansible.module_utils.common.text.converters import to_native
import base64
import hashlib
import io
import json
import os
import tempfile
import time


class FailedInstallingWithPluginManager(Exception):
    pass


class JenkinsPlugin(object):
    def __init__(self, module):
        # To be able to call fail_json
        self.module = module

        # Shortcuts for the params
        self.params = self.module.params
        self.url = self.params['url']
        self.timeout = self.params['timeout']

        # Crumb
        self.crumb = {}
        # Cookie jar for crumb session
        self.cookies = None

        if self._csrf_enabled():
            self.cookies = cookiejar.LWPCookieJar()
            self.crumb = self._get_crumb()

        # Get list of installed plugins
        self._get_installed_plugins()

    def _csrf_enabled(self):
        csrf_data = self._get_json_data(
            "%s/%s" % (self.url, "api/json"), 'CSRF')

        if 'useCrumbs' not in csrf_data:
            self.module.fail_json(
                msg="Required fields not found in the Crumbs response.",
                details=csrf_data)

        return csrf_data['useCrumbs']

    def _get_json_data(self, url, what, **kwargs):
        # Get the JSON data
        r = self._get_url_data(url, what, **kwargs)

        # Parse the JSON data
        try:
            json_data = json.loads(to_native(r.read()))
        except Exception as e:
            self.module.fail_json(
                msg="Cannot parse %s JSON data." % what,
                details=to_native(e))

        return json_data

    def _get_urls_data(self, urls, what=None, msg_status=None, msg_exception=None, **kwargs):
        # Compose default messages
        if msg_status is None:
            msg_status = "Cannot get %s" % what

        if msg_exception is None:
            msg_exception = "Retrieval of %s failed." % what

        errors = {}
        for url in urls:
            err_msg = None
            try:
                self.module.debug("fetching url: %s" % url)
                response, info = fetch_url(
                    self.module, url, timeout=self.timeout, cookies=self.cookies,
                    headers=self.crumb, **kwargs)

                if info['status'] == 200:
                    return response
                else:
                    err_msg = ("%s. fetching url %s failed. response code: %s" % (msg_status, url, info['status']))
                    if info['status'] > 400:  # extend error message
                        err_msg = "%s. response body: %s" % (err_msg, info['body'])
            except Exception as e:
                err_msg = "%s. fetching url %s failed. error msg: %s" % (msg_status, url, to_native(e))
            finally:
                if err_msg is not None:
                    self.module.debug(err_msg)
                    errors[url] = err_msg

        # failed on all urls
        self.module.fail_json(msg=msg_exception, details=errors)

    def _get_url_data(
            self, url, what=None, msg_status=None, msg_exception=None,
            dont_fail=False, **kwargs):
        # Compose default messages
        if msg_status is None:
            msg_status = "Cannot get %s" % what

        if msg_exception is None:
            msg_exception = "Retrieval of %s failed." % what

        # Get the URL data
        try:
            response, info = fetch_url(
                self.module, url, timeout=self.timeout, cookies=self.cookies,
                headers=self.crumb, **kwargs)

            if info['status'] != 200:
                if dont_fail:
                    raise FailedInstallingWithPluginManager(info['msg'])
                else:
                    self.module.fail_json(msg=msg_status, details=info['msg'])
        except Exception as e:
            if dont_fail:
                raise FailedInstallingWithPluginManager(e)
            else:
                self.module.fail_json(msg=msg_exception, details=to_native(e))

        return response

    def _get_crumb(self):
        crumb_data = self._get_json_data(
            "%s/%s" % (self.url, "crumbIssuer/api/json"), 'Crumb')

        if 'crumbRequestField' in crumb_data and 'crumb' in crumb_data:
            ret = {
                crumb_data['crumbRequestField']: crumb_data['crumb']
            }
        else:
            self.module.fail_json(
                msg="Required fields not found in the Crum response.",
                details=crumb_data)

        return ret

    def _get_installed_plugins(self):
        plugins_data = self._get_json_data(
            "%s/%s" % (self.url, "pluginManager/api/json?depth=1"),
            'list of plugins')

        # Check if we got valid data
        if 'plugins' not in plugins_data:
            self.module.fail_json(msg="No valid plugin data found.")

        # Create final list of installed/pined plugins
        self.is_installed = False
        self.is_pinned = False
        self.is_enabled = False

        for p in plugins_data['plugins']:
            if p['shortName'] == self.params['name']:
                self.is_installed = True

                if p['pinned']:
                    self.is_pinned = True

                if p['enabled']:
                    self.is_enabled = True

                break

    def _install_with_plugin_manager(self):
        if not self.module.check_mode:
            # Install the plugin (with dependencies)
            install_script = (
                'd = Jenkins.instance.updateCenter.getPlugin("%s")'
                '.deploy(); d.get();' % self.params['name'])

            if self.params['with_dependencies']:
                install_script = (
                    'Jenkins.instance.updateCenter.getPlugin("%s")'
                    '.getNeededDependencies().each{it.deploy()}; %s' % (
                        self.params['name'], install_script))

            script_data = {
                'script': install_script
            }
            data = urlencode(script_data)

            # Send the installation request
            r = self._get_url_data(
                "%s/scriptText" % self.url,
                msg_status="Cannot install plugin.",
                msg_exception="Plugin installation has failed.",
                data=data,
                dont_fail=True)

            hpi_file = '%s/plugins/%s.hpi' % (
                self.params['jenkins_home'],
                self.params['name'])

            if os.path.isfile(hpi_file):
                os.remove(hpi_file)

    def install(self):
        changed = False
        plugin_file = (
            '%s/plugins/%s.jpi' % (
                self.params['jenkins_home'],
                self.params['name']))

        if not self.is_installed and self.params['version'] in [None, 'latest']:
            try:
                self._install_with_plugin_manager()
                changed = True
            except FailedInstallingWithPluginManager:  # Fallback to manually downloading the plugin
                pass

        if not changed:
            # Check if the plugin directory exists
            if not os.path.isdir(self.params['jenkins_home']):
                self.module.fail_json(
                    msg="Jenkins home directory doesn't exist.")

            checksum_old = None
            if os.path.isfile(plugin_file):
                # Make the checksum of the currently installed plugin
                with open(plugin_file, 'rb') as plugin_fh:
                    plugin_content = plugin_fh.read()
                checksum_old = hashlib.sha1(plugin_content).hexdigest()

            if self.params['version'] in [None, 'latest']:
                # Take latest version
                plugin_urls = self._get_latest_plugin_urls()
            else:
                # Take specific version
                plugin_urls = self._get_versioned_plugin_urls()
            if (
                    self.params['updates_expiration'] == 0 or
                    self.params['version'] not in [None, 'latest'] or
                    checksum_old is None):

                # Download the plugin file directly
                r = self._download_plugin(plugin_urls)

                # Write downloaded plugin into file if checksums don't match
                if checksum_old is None:
                    # No previously installed plugin
                    if not self.module.check_mode:
                        self._write_file(plugin_file, r)

                    changed = True
                else:
                    # Get data for the MD5
                    data = r.read()

                    # Make new checksum
                    checksum_new = hashlib.sha1(data).hexdigest()

                    # If the checksum is different from the currently installed
                    # plugin, store the new plugin
                    if checksum_old != checksum_new:
                        if not self.module.check_mode:
                            self._write_file(plugin_file, data)

                        changed = True
            elif self.params['version'] == 'latest':
                # Check for update from the updates JSON file
                plugin_data = self._download_updates()

                # If the latest version changed, download it
                if checksum_old != to_bytes(plugin_data['sha1']):
                    if not self.module.check_mode:
                        r = self._download_plugin(plugin_urls)
                        self._write_file(plugin_file, r)

                    changed = True

        # Change file attributes if needed
        if os.path.isfile(plugin_file):
            params = {
                'dest': plugin_file
            }
            params.update(self.params)
            file_args = self.module.load_file_common_arguments(params)

            if not self.module.check_mode:
                # Not sure how to run this in the check mode
                changed = self.module.set_fs_attributes_if_different(
                    file_args, changed)
            else:
                # See the comment above
                changed = True

        return changed

    def _get_latest_plugin_urls(self):
        urls = []
        for base_url in self.params['updates_url']:
            for update_segment in self.params['latest_plugins_url_segments']:
                urls.append("{0}/{1}/{2}.hpi".format(base_url, update_segment, self.params['name']))
        return urls

    def _get_versioned_plugin_urls(self):
        urls = []
        for base_url in self.params['updates_url']:
            for versioned_segment in self.params['versioned_plugins_url_segments']:
                urls.append("{0}/{1}/{2}/{3}/{2}.hpi".format(base_url, versioned_segment, self.params['name'], self.params['version']))
        return urls

    def _get_update_center_urls(self):
        urls = []
        for base_url in self.params['updates_url']:
            for update_json in self.params['update_json_url_segment']:
                urls.append("{0}/{1}".format(base_url, update_json))
        return urls

    def _download_updates(self):
        updates_filename = 'jenkins-plugin-cache.json'
        updates_dir = os.path.expanduser('~/.ansible/tmp')
        updates_file = "%s/%s" % (updates_dir, updates_filename)
        download_updates = True

        # Check if we need to download new updates file
        if os.path.isfile(updates_file):
            # Get timestamp when the file was changed last time
            ts_file = os.stat(updates_file).st_mtime
            ts_now = time.time()

            if ts_now - ts_file < self.params['updates_expiration']:
                download_updates = False

        updates_file_orig = updates_file

        # Download the updates file if needed
        if download_updates:
            urls = self._get_update_center_urls()

            # Get the data
            r = self._get_urls_data(
                urls,
                msg_status="Remote updates not found.",
                msg_exception="Updates download failed.")

            # Write the updates file
            update_fd, updates_file = tempfile.mkstemp()
            os.write(update_fd, r.read())

            try:
                os.close(update_fd)
            except IOError as e:
                self.module.fail_json(
                    msg="Cannot close the tmp updates file %s." % updates_file,
                    details=to_native(e))

        # Open the updates file
        try:
            f = io.open(updates_file, encoding='utf-8')
        except IOError as e:
            self.module.fail_json(
                msg="Cannot open temporal updates file.",
                details=to_native(e))

        i = 0
        for line in f:
            # Read only the second line
            if i == 1:
                try:
                    data = json.loads(line)
                except Exception as e:
                    self.module.fail_json(
                        msg="Cannot load JSON data from the tmp updates file.",
                        details=to_native(e))

                break

            i += 1

        # Move the updates file to the right place if we could read it
        if download_updates:
            # Make sure the destination directory exists
            if not os.path.isdir(updates_dir):
                try:
                    os.makedirs(updates_dir, int('0700', 8))
                except OSError as e:
                    self.module.fail_json(
                        msg="Cannot create temporal directory.",
                        details=to_native(e))

            self.module.atomic_move(updates_file, updates_file_orig)

        # Check if we have the plugin data available
        if 'plugins' not in data or self.params['name'] not in data['plugins']:
            self.module.fail_json(
                msg="Cannot find plugin data in the updates file.")

        return data['plugins'][self.params['name']]

    def _download_plugin(self, plugin_urls):
        # Download the plugin

        return self._get_urls_data(
            plugin_urls,
            msg_status="Plugin not found.",
            msg_exception="Plugin download failed.")

    def _write_file(self, f, data):
        # Store the plugin into a temp file and then move it
        tmp_f_fd, tmp_f = tempfile.mkstemp()

        if isinstance(data, (text_type, binary_type)):
            os.write(tmp_f_fd, data)
        else:
            os.write(tmp_f_fd, data.read())

        try:
            os.close(tmp_f_fd)
        except IOError as e:
            self.module.fail_json(
                msg='Cannot close the temporal plugin file %s.' % tmp_f,
                details=to_native(e))

        # Move the file onto the right place
        self.module.atomic_move(tmp_f, f)

    def uninstall(self):
        changed = False

        # Perform the action
        if self.is_installed:
            if not self.module.check_mode:
                self._pm_query('doUninstall', 'Uninstallation')

            changed = True

        return changed

    def pin(self):
        return self._pinning('pin')

    def unpin(self):
        return self._pinning('unpin')

    def _pinning(self, action):
        changed = False

        # Check if the plugin is pinned/unpinned
        if (
                action == 'pin' and not self.is_pinned or
                action == 'unpin' and self.is_pinned):

            # Perform the action
            if not self.module.check_mode:
                self._pm_query(action, "%sning" % action.capitalize())

            changed = True

        return changed

    def enable(self):
        return self._enabling('enable')

    def disable(self):
        return self._enabling('disable')

    def _enabling(self, action):
        changed = False

        # Check if the plugin is pinned/unpinned
        if (
                action == 'enable' and not self.is_enabled or
                action == 'disable' and self.is_enabled):

            # Perform the action
            if not self.module.check_mode:
                self._pm_query(
                    "make%sd" % action.capitalize(),
                    "%sing" % action[:-1].capitalize())

            changed = True

        return changed

    def _pm_query(self, action, msg):
        url = "%s/pluginManager/plugin/%s/%s" % (
            self.params['url'], self.params['name'], action)

        # Send the request
        self._get_url_data(
            url,
            msg_status="Plugin not found. %s" % url,
            msg_exception="%s has failed." % msg,
            method="POST")


def main():
    # Module arguments
    argument_spec = url_argument_spec()
    argument_spec.update(
        group=dict(type='str', default='jenkins'),
        jenkins_home=dict(type='path', default='/var/lib/jenkins'),
        mode=dict(default='0644', type='raw'),
        name=dict(type='str', required=True),
        owner=dict(type='str', default='jenkins'),
        state=dict(
            choices=[
                'present',
                'absent',
                'pinned',
                'unpinned',
                'enabled',
                'disabled',
                'latest'],
            default='present'),
        timeout=dict(default=30, type="int"),
        updates_expiration=dict(default=86400, type="int"),
        updates_url=dict(type="list", elements="str", default=['https://updates.jenkins.io',
                                                               'http://mirrors.jenkins.io']),
        update_json_url_segment=dict(type="list", elements="str", default=['update-center.json',
                                                                           'updates/update-center.json']),
        latest_plugins_url_segments=dict(type="list", elements="str", default=['latest']),
        versioned_plugins_url_segments=dict(type="list", elements="str", default=['download/plugins', 'plugins']),
        url=dict(default='http://localhost:8080'),
        url_password=dict(no_log=True),
        version=dict(),
        with_dependencies=dict(default=True, type='bool'),
    )
    # Module settings
    module = AnsibleModule(
        argument_spec=argument_spec,
        add_file_common_args=True,
        supports_check_mode=True,
    )

    # Force basic authentication
    module.params['force_basic_auth'] = True

    # Convert timeout to float
    try:
        module.params['timeout'] = float(module.params['timeout'])
    except ValueError as e:
        module.fail_json(
            msg='Cannot convert %s to float.' % module.params['timeout'],
            details=to_native(e))

    # Set version to latest if state is latest
    if module.params['state'] == 'latest':
        module.params['state'] = 'present'
        module.params['version'] = 'latest'

    # Create some shortcuts
    name = module.params['name']
    state = module.params['state']

    # Initial change state of the task
    changed = False

    # Instantiate the JenkinsPlugin object
    jp = JenkinsPlugin(module)

    # Perform action depending on the requested state
    if state == 'present':
        changed = jp.install()
    elif state == 'absent':
        changed = jp.uninstall()
    elif state == 'pinned':
        changed = jp.pin()
    elif state == 'unpinned':
        changed = jp.unpin()
    elif state == 'enabled':
        changed = jp.enable()
    elif state == 'disabled':
        changed = jp.disable()

    # Print status of the change
    module.exit_json(changed=changed, plugin=name, state=state)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: profitbricks_nic
short_description: Create or Remove a NIC.
description:
     - This module allows you to create or restore a volume snapshot. This module has a dependency on profitbricks >= 1.0.0
options:
  datacenter:
    description:
      - The datacenter in which to operate.
    type: str
    required: true
  server:
    description:
      - The server name or ID.
    type: str
    required: true
  name:
    description:
      - The name or ID of the NIC. This is only required on deletes, but not on create.
      - If not specified, it defaults to a value based on UUID4.
    type: str
  lan:
    description:
      - The LAN to place the NIC on. You can pass a LAN that doesn't exist and it will be created. Required on create.
    type: str
  subscription_user:
    description:
      - The ProfitBricks username. Overrides the PB_SUBSCRIPTION_ID environment variable.
    type: str
    required: true
  subscription_password:
    description:
      - THe ProfitBricks password. Overrides the PB_PASSWORD environment variable.
    type: str
    required: true
  wait:
    description:
      - wait for the operation to complete before returning
    required: false
    default: "yes"
    type: bool
  wait_timeout:
    description:
      - how long before wait gives up, in seconds
    type: int
    default: 600
  state:
    description:
      - Indicate desired state of the resource
      - "The available choices are: C(present), C(absent)."
    type: str
    required: false
    default: 'present'

requirements: [ "profitbricks" ]
author: Matt Baldwin (@baldwinSPC) <baldwin@stackpointcloud.com>
'''

EXAMPLES = '''
- name: Create a NIC
  community.general.profitbricks_nic:
    datacenter: Tardis One
    server: node002
    lan: 2
    wait_timeout: 500
    state: present

- name: Remove a NIC
  community.general.profitbricks_nic:
    datacenter: Tardis One
    server: node002
    name: 7341c2454f
    wait_timeout: 500
    state: absent
'''

import re
import uuid
import time

HAS_PB_SDK = True
try:
    from profitbricks.client import ProfitBricksService, NIC
except ImportError:
    HAS_PB_SDK = False

from ansible.module_utils.basic import AnsibleModule


uuid_match = re.compile(
    r'[\w]{8}-[\w]{4}-[\w]{4}-[\w]{4}-[\w]{12}', re.I)


def _make_default_name():
    return str(uuid.uuid4()).replace('-', '')[:10]


def _wait_for_completion(profitbricks, promise, wait_timeout, msg):
    if not promise:
        return
    wait_timeout = time.time() + wait_timeout
    while wait_timeout > time.time():
        time.sleep(5)
        operation_result = profitbricks.get_request(
            request_id=promise['requestId'],
            status=True)

        if operation_result['metadata']['status'] == "DONE":
            return
        elif operation_result['metadata']['status'] == "FAILED":
            raise Exception(
                'Request failed to complete ' + msg + ' "' + str(
                    promise['requestId']) + '" to complete.')

    raise Exception(
        'Timed out waiting for async operation ' + msg + ' "' + str(
            promise['requestId']
        ) + '" to complete.')


def create_nic(module, profitbricks):
    """
    Creates a NIC.

    module : AnsibleModule object
    profitbricks: authenticated profitbricks object.

    Returns:
        True if the nic creates, false otherwise
    """
    datacenter = module.params.get('datacenter')
    server = module.params.get('server')
    lan = module.params.get('lan')
    name = module.params.get('name')
    if name is None:
        name = _make_default_name()
    wait = module.params.get('wait')
    wait_timeout = module.params.get('wait_timeout')

    # Locate UUID for Datacenter
    if not (uuid_match.match(datacenter)):
        datacenter_list = profitbricks.list_datacenters()
        for d in datacenter_list['items']:
            dc = profitbricks.get_datacenter(d['id'])
            if datacenter == dc['properties']['name']:
                datacenter = d['id']
                break

    # Locate UUID for Server
    if not (uuid_match.match(server)):
        server_list = profitbricks.list_servers(datacenter)
        for s in server_list['items']:
            if server == s['properties']['name']:
                server = s['id']
                break
    try:
        n = NIC(
            name=name,
            lan=lan
        )

        nic_response = profitbricks.create_nic(datacenter, server, n)

        if wait:
            _wait_for_completion(profitbricks, nic_response,
                                 wait_timeout, "create_nic")

        return nic_response

    except Exception as e:
        module.fail_json(msg="failed to create the NIC: %s" % str(e))


def delete_nic(module, profitbricks):
    """
    Removes a NIC

    module : AnsibleModule object
    profitbricks: authenticated profitbricks object.

    Returns:
        True if the NIC was removed, false otherwise
    """
    datacenter = module.params.get('datacenter')
    server = module.params.get('server')
    name = module.params.get('name')
    if name is None:
        name = _make_default_name()

    # Locate UUID for Datacenter
    if not (uuid_match.match(datacenter)):
        datacenter_list = profitbricks.list_datacenters()
        for d in datacenter_list['items']:
            dc = profitbricks.get_datacenter(d['id'])
            if datacenter == dc['properties']['name']:
                datacenter = d['id']
                break

    # Locate UUID for Server
    server_found = False
    if not (uuid_match.match(server)):
        server_list = profitbricks.list_servers(datacenter)
        for s in server_list['items']:
            if server == s['properties']['name']:
                server_found = True
                server = s['id']
                break

        if not server_found:
            return False

    # Locate UUID for NIC
    nic_found = False
    if not (uuid_match.match(name)):
        nic_list = profitbricks.list_nics(datacenter, server)
        for n in nic_list['items']:
            if name == n['properties']['name']:
                nic_found = True
                name = n['id']
                break

        if not nic_found:
            return False

    try:
        nic_response = profitbricks.delete_nic(datacenter, server, name)
        return nic_response
    except Exception as e:
        module.fail_json(msg="failed to remove the NIC: %s" % str(e))


def main():
    module = AnsibleModule(
        argument_spec=dict(
            datacenter=dict(required=True),
            server=dict(required=True),
            name=dict(),
            lan=dict(),
            subscription_user=dict(required=True),
            subscription_password=dict(required=True, no_log=True),
            wait=dict(type='bool', default=True),
            wait_timeout=dict(type='int', default=600),
            state=dict(default='present'),
        ),
        required_if=(
            ('state', 'absent', ['name']),
            ('state', 'present', ['lan']),
        )
    )

    if not HAS_PB_SDK:
        module.fail_json(msg='profitbricks required for this module')

    subscription_user = module.params.get('subscription_user')
    subscription_password = module.params.get('subscription_password')

    profitbricks = ProfitBricksService(
        username=subscription_user,
        password=subscription_password)

    state = module.params.get('state')

    if state == 'absent':
        try:
            (changed) = delete_nic(module, profitbricks)
            module.exit_json(changed=changed)
        except Exception as e:
            module.fail_json(msg='failed to set nic state: %s' % str(e))

    elif state == 'present':
        try:
            (nic_dict) = create_nic(module, profitbricks)
            module.exit_json(nics=nic_dict)  # @FIXME changed not calculated?
        except Exception as e:
            module.fail_json(msg='failed to set nic state: %s' % str(e))


if __name__ == '__main__':
    main()



# Author: Alexei Znamensky (russoz@gmail.com)
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import json

from ansible_collections.community.general.plugins.modules.system import xfconf_info

import pytest

TESTED_MODULE = xfconf_info.__name__


@pytest.fixture
def patch_xfconf_info(mocker):
    """
    Function used for mocking some parts of redhat_subscribtion module
    """
    mocker.patch('ansible_collections.community.general.plugins.module_utils.mh.module_helper.AnsibleModule.get_bin_path',
                 return_value='/testbin/xfconf-query')


TEST_CASES = [
    [
        {'channel': 'xfwm4', 'property': '/general/inactive_opacity'},
        {
            'id': 'test_simple_property_get',
            'run_command.calls': [
                (
                    # Calling of following command will be asserted
                    ['/testbin/xfconf-query', '--channel', 'xfwm4', '--property', '/general/inactive_opacity'],
                    # Was return code checked?
                    {'environ_update': {'LANGUAGE': 'C', 'LC_ALL': 'C'}, 'check_rc': True},
                    # Mock of returned code, stdout and stderr
                    (0, '100\n', '',),
                ),
            ],
            'is_array': False,
            'value': '100',
        }
    ],
    [
        {'channel': 'xfwm4', 'property': '/general/i_dont_exist'},
        {
            'id': 'test_simple_property_get_nonexistent',
            'run_command.calls': [
                (
                    # Calling of following command will be asserted
                    ['/testbin/xfconf-query', '--channel', 'xfwm4', '--property', '/general/i_dont_exist'],
                    # Was return code checked?
                    {'environ_update': {'LANGUAGE': 'C', 'LC_ALL': 'C'}, 'check_rc': True},
                    # Mock of returned code, stdout and stderr
                    (1, '', 'Property "/general/i_dont_exist" does not exist on channel "xfwm4".\n',),
                ),
            ],
            'is_array': False,
        }
    ],
    [
        {'property': '/general/i_dont_exist'},
        {
            'id': 'test_property_no_channel',
            'run_command.calls': [],
        }
    ],
    [
        {'channel': 'xfwm4', 'property': '/general/workspace_names'},
        {
            'id': 'test_property_get_array',
            'run_command.calls': [
                (
                    # Calling of following command will be asserted
                    ['/testbin/xfconf-query', '--channel', 'xfwm4', '--property', '/general/workspace_names'],
                    # Was return code checked?
                    {'environ_update': {'LANGUAGE': 'C', 'LC_ALL': 'C'}, 'check_rc': True},
                    # Mock of returned code, stdout and stderr
                    (0, 'Value is an array with 3 items:\n\nMain\nWork\nTmp\n', '',),
                ),
            ],
            'is_array': True,
            'value_array': ['Main', 'Work', 'Tmp'],
        },
    ],
    [
        {},
        {
            'id': 'get_channels',
            'run_command.calls': [
                (
                    # Calling of following command will be asserted
                    ['/testbin/xfconf-query', '--list'],
                    # Was return code checked?
                    {'environ_update': {'LANGUAGE': 'C', 'LC_ALL': 'C'}, 'check_rc': True},
                    # Mock of returned code, stdout and stderr
                    (0, 'Channels:\n  a\n  b\n  c\n', '',),
                ),
            ],
            'is_array': False,
            'channels': ['a', 'b', 'c'],
        },
    ],
    [
        {'channel': 'xfwm4'},
        {
            'id': 'get_properties',
            'run_command.calls': [
                (
                    # Calling of following command will be asserted
                    ['/testbin/xfconf-query', '--list', '--channel', 'xfwm4'],
                    # Was return code checked?
                    {'environ_update': {'LANGUAGE': 'C', 'LC_ALL': 'C'}, 'check_rc': True},
                    # Mock of returned code, stdout and stderr
                    (0, '/general/wrap_cycle\n/general/wrap_layout\n/general/wrap_resistance\n/general/wrap_windows\n'
                        '/general/wrap_workspaces\n/general/zoom_desktop\n', '',),
                ),
            ],
            'is_array': False,
            'properties': [
                '/general/wrap_cycle',
                '/general/wrap_layout',
                '/general/wrap_resistance',
                '/general/wrap_windows',
                '/general/wrap_workspaces',
                '/general/zoom_desktop',
            ],
        },
    ],
]
TEST_CASES_IDS = [item[1]['id'] for item in TEST_CASES]


@pytest.mark.parametrize('patch_ansible_module, testcase',
                         TEST_CASES,
                         ids=TEST_CASES_IDS,
                         indirect=['patch_ansible_module'])
@pytest.mark.usefixtures('patch_ansible_module')
def test_xfconf_info(mocker, capfd, patch_xfconf_info, testcase):
    """
    Run unit tests for test cases listen in TEST_CASES
    """

    # Mock function used for running commands first
    call_results = [item[2] for item in testcase['run_command.calls']]
    mock_run_command = mocker.patch(
        'ansible_collections.community.general.plugins.module_utils.mh.module_helper.AnsibleModule.run_command',
        side_effect=call_results)

    # Try to run test case
    with pytest.raises(SystemExit):
        xfconf_info.main()

    out, err = capfd.readouterr()
    results = json.loads(out)
    print("testcase =\n%s" % testcase)
    print("results =\n%s" % results)

    for conditional_test_result in ('value_array', 'value', 'is_array', 'properties', 'channels'):
        if conditional_test_result in testcase:
            assert conditional_test_result in results, "'{0}' not found in {1}".format(conditional_test_result, results)
            assert results[conditional_test_result] == testcase[conditional_test_result], \
                "'{0}': '{1}' != '{2}'".format(conditional_test_result, results[conditional_test_result], testcase[conditional_test_result])

    assert mock_run_command.call_count == len(testcase['run_command.calls'])
    if mock_run_command.call_count:
        call_args_list = [(item[0][0], item[1]) for item in mock_run_command.call_args_list]
        expected_call_args_list = [(item[0], item[1]) for item in testcase['run_command.calls']]
        print("call args list =\n%s" % call_args_list)
        print("expected args list =\n%s" % expected_call_args_list)
        assert call_args_list == expected_call_args_list



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) Vincent Van de Kussen
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: rhn_channel
short_description: Adds or removes Red Hat software channels
description:
    - Adds or removes Red Hat software channels.
author:
- Vincent Van der Kussen (@vincentvdk)
notes:
    - This module fetches the system id from RHN.
    - This module doesn't support I(check_mode).
options:
    name:
        description:
            - Name of the software channel.
        required: true
        type: str
    sysname:
        description:
            - Name of the system as it is known in RHN/Satellite.
        required: true
        type: str
    state:
        description:
            - Whether the channel should be present or not, taking action if the state is different from what is stated.
        default: present
        choices: [ present, absent ]
        type: str
    url:
        description:
            - The full URL to the RHN/Satellite API.
        required: true
        type: str
    user:
        description:
            - RHN/Satellite login.
        required: true
        type: str
    password:
        description:
            - RHN/Satellite password.
        aliases: [pwd]
        required: true
        type: str
    validate_certs:
        description:
            - If C(False), SSL certificates will not be validated.
            - This should only set to C(False) when used on self controlled sites
              using self-signed certificates, and you are absolutely sure that nobody
              can modify traffic between the module and the site.
        type: bool
        default: true
        version_added: '0.2.0'
'''

EXAMPLES = '''
- name: Add a Red Hat software channel
  community.general.rhn_channel:
    name: rhel-x86_64-server-v2vwin-6
    sysname: server01
    url: https://rhn.redhat.com/rpc/api
    user: rhnuser
    password: guessme
  delegate_to: localhost
'''

import ssl
from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.six.moves import xmlrpc_client


def get_systemid(client, session, sysname):
    systems = client.system.listUserSystems(session)
    for system in systems:
        if system.get('name') == sysname:
            idres = system.get('id')
            idd = int(idres)
            return idd


def subscribe_channels(channelname, client, session, sysname, sys_id):
    channels = base_channels(client, session, sys_id)
    channels.append(channelname)
    return client.system.setChildChannels(session, sys_id, channels)


def unsubscribe_channels(channelname, client, session, sysname, sys_id):
    channels = base_channels(client, session, sys_id)
    channels.remove(channelname)
    return client.system.setChildChannels(session, sys_id, channels)


def base_channels(client, session, sys_id):
    basechan = client.channel.software.listSystemChannels(session, sys_id)
    try:
        chans = [item['label'] for item in basechan]
    except KeyError:
        chans = [item['channel_label'] for item in basechan]
    return chans


def main():

    module = AnsibleModule(
        argument_spec=dict(
            state=dict(type='str', default='present', choices=['present', 'absent']),
            name=dict(type='str', required=True),
            sysname=dict(type='str', required=True),
            url=dict(type='str', required=True),
            user=dict(type='str', required=True),
            password=dict(type='str', required=True, aliases=['pwd'], no_log=True),
            validate_certs=dict(type='bool', default=True),
        )
    )

    state = module.params['state']
    channelname = module.params['name']
    systname = module.params['sysname']
    saturl = module.params['url']
    user = module.params['user']
    password = module.params['password']
    validate_certs = module.params['validate_certs']

    ssl_context = None
    if not validate_certs:
        try:  # Python 2.7.9 and newer
            ssl_context = ssl.create_unverified_context()
        except AttributeError:  # Legacy Python that doesn't verify HTTPS certificates by default
            ssl_context = ssl._create_unverified_context()
        else:  # Python 2.7.8 and older
            ssl._create_default_https_context = ssl._create_unverified_https_context

    # initialize connection
    if ssl_context:
        client = xmlrpc_client.ServerProxy(saturl, context=ssl_context)
    else:
        client = xmlrpc_client.Server(saturl)

    try:
        session = client.auth.login(user, password)
    except Exception as e:
        module.fail_json(msg="Unable to establish session with Satellite server: %s " % to_text(e))

    if not session:
        module.fail_json(msg="Failed to establish session with Satellite server.")

    # get systemid
    try:
        sys_id = get_systemid(client, session, systname)
    except Exception as e:
        module.fail_json(msg="Unable to get system id: %s " % to_text(e))

    if not sys_id:
        module.fail_json(msg="Failed to get system id.")

    # get channels for system
    try:
        chans = base_channels(client, session, sys_id)
    except Exception as e:
        module.fail_json(msg="Unable to get channel information: %s " % to_text(e))

    try:
        if state == 'present':
            if channelname in chans:
                module.exit_json(changed=False, msg="Channel %s already exists" % channelname)
            else:
                subscribe_channels(channelname, client, session, systname, sys_id)
                module.exit_json(changed=True, msg="Channel %s added" % channelname)

        if state == 'absent':
            if channelname not in chans:
                module.exit_json(changed=False, msg="Not subscribed to channel %s." % channelname)
            else:
                unsubscribe_channels(channelname, client, session, systname, sys_id)
                module.exit_json(changed=True, msg="Channel %s removed" % channelname)
    except Exception as e:
        module.fail_json(msg='Unable to %s channel (%s): %s' % ('add' if state == 'present' else 'remove', channelname, to_text(e)))
    finally:
        client.auth.logout(session)


if __name__ == '__main__':
    main()



# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible_collections.community.general.tests.unit.compat import unittest
from ansible_collections.community.general.tests.unit.compat.mock import patch
from ansible.module_utils import basic
from ansible.module_utils.common.text.converters import to_bytes
from ansible_collections.community.general.plugins.modules.web_infrastructure import jenkins_build

import json


def set_module_args(args):
    """prepare arguments so that they will be picked up during module creation"""
    args = json.dumps({'ANSIBLE_MODULE_ARGS': args})
    basic._ANSIBLE_ARGS = to_bytes(args)


class AnsibleExitJson(Exception):
    """Exception class to be raised by module.exit_json and caught by the test case"""
    pass


class AnsibleFailJson(Exception):
    """Exception class to be raised by module.fail_json and caught by the test case"""
    pass


def exit_json(*args, **kwargs):
    """function to patch over exit_json; package return data into an exception"""
    if 'changed' not in kwargs:
        kwargs['changed'] = False
    raise AnsibleExitJson(kwargs)


def fail_json(*args, **kwargs):
    """function to patch over fail_json; package return data into an exception"""
    kwargs['failed'] = True
    raise AnsibleFailJson(kwargs)


class JenkinsMock():

    def get_job_info(self, name):
        return {
            "nextBuildNumber": 1234
        }

    def get_build_info(self, name, build_number):
        return {
            "building": True,
            "result": "SUCCESS"
        }

    def build_job(self, *args):
        return None

    def delete_build(self, name, build_number):
        return None

    def stop_build(self, name, build_number):
        return None


class JenkinsMockIdempotent():

    def get_job_info(self, name):
        return {
            "nextBuildNumber": 1235
        }

    def get_build_info(self, name, build_number):
        return {
            "building": False,
            "result": "ABORTED"
        }

    def build_job(self, *args):
        return None

    def delete_build(self, name, build_number):
        return None

    def stop_build(self, name, build_number):
        return None


class TestJenkinsBuild(unittest.TestCase):

    def setUp(self):
        self.mock_module_helper = patch.multiple(basic.AnsibleModule,
                                                 exit_json=exit_json,
                                                 fail_json=fail_json)
        self.mock_module_helper.start()
        self.addCleanup(self.mock_module_helper.stop)

    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.test_dependencies')
    def test_module_fail_when_required_args_missing(self, test_deps):
        test_deps.return_value = None
        with self.assertRaises(AnsibleFailJson):
            set_module_args({})
            jenkins_build.main()

    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.test_dependencies')
    def test_module_fail_when_missing_build_number(self, test_deps):
        test_deps.return_value = None
        with self.assertRaises(AnsibleFailJson):
            set_module_args({
                "name": "required-if",
                "state": "stopped"
            })
            jenkins_build.main()

    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.test_dependencies')
    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.JenkinsBuild.get_jenkins_connection')
    def test_module_create_build(self, jenkins_connection, test_deps):
        test_deps.return_value = None
        jenkins_connection.return_value = JenkinsMock()

        with self.assertRaises(AnsibleExitJson):
            set_module_args({
                "name": "host-check",
                "user": "abc",
                "token": "xyz"
            })
            jenkins_build.main()

    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.test_dependencies')
    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.JenkinsBuild.get_jenkins_connection')
    def test_module_stop_build(self, jenkins_connection, test_deps):
        test_deps.return_value = None
        jenkins_connection.return_value = JenkinsMock()

        with self.assertRaises(AnsibleExitJson) as return_json:
            set_module_args({
                "name": "host-check",
                "build_number": "1234",
                "state": "stopped",
                "user": "abc",
                "token": "xyz"
            })
            jenkins_build.main()

        self.assertTrue(return_json.exception.args[0]['changed'])

    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.test_dependencies')
    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.JenkinsBuild.get_jenkins_connection')
    def test_module_stop_build_again(self, jenkins_connection, test_deps):
        test_deps.return_value = None
        jenkins_connection.return_value = JenkinsMockIdempotent()

        with self.assertRaises(AnsibleExitJson) as return_json:
            set_module_args({
                "name": "host-check",
                "build_number": "1234",
                "state": "stopped",
                "user": "abc",
                "password": "xyz"
            })
            jenkins_build.main()

        self.assertFalse(return_json.exception.args[0]['changed'])

    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.test_dependencies')
    @patch('ansible_collections.community.general.plugins.modules.web_infrastructure.jenkins_build.JenkinsBuild.get_jenkins_connection')
    def test_module_delete_build(self, jenkins_connection, test_deps):
        test_deps.return_value = None
        jenkins_connection.return_value = JenkinsMock()

        with self.assertRaises(AnsibleExitJson):
            set_module_args({
                "name": "host-check",
                "build_number": "1234",
                "state": "absent",
                "user": "abc",
                "token": "xyz"
            })
            jenkins_build.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2020-2022, Vladimir Botka <vbotka@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.errors import AnsibleFilterError
from ansible.module_utils.six import string_types
from ansible.module_utils.common._collections_compat import Mapping, Sequence
from ansible.utils.vars import merge_hash
from ansible.release import __version__ as ansible_version
from ansible_collections.community.general.plugins.module_utils.version import LooseVersion

from collections import defaultdict
from operator import itemgetter


def merge_hash_wrapper(x, y, recursive=False, list_merge='replace'):
    ''' Wrapper of the function merge_hash from ansible.utils.vars. Only 2 paramaters are allowed
        for Ansible 2.9 and lower.'''

    if LooseVersion(ansible_version) < LooseVersion('2.10'):
        if list_merge != 'replace' or recursive:
            msg = ("Non default options of list_merge(default=replace) or recursive(default=False) "
                   "are not allowed in Ansible version 2.9 or lower. Ansible version is %s, "
                   "recursive=%s, and list_merge=%s.")
            raise AnsibleFilterError(msg % (ansible_version, recursive, list_merge))
        else:
            return merge_hash(x, y)
    else:
        return merge_hash(x, y, recursive, list_merge)


def list_mergeby(x, y, index, recursive=False, list_merge='replace'):
    ''' Merge 2 lists by attribute 'index'. The function merge_hash from ansible.utils.vars is used.
        This function is used by the function lists_mergeby.
    '''

    d = defaultdict(dict)
    for l in (x, y):
        for elem in l:
            if not isinstance(elem, Mapping):
                msg = "Elements of list arguments for lists_mergeby must be dictionaries. %s is %s"
                raise AnsibleFilterError(msg % (elem, type(elem)))
            if index in elem.keys():
                d[elem[index]].update(merge_hash_wrapper(d[elem[index]], elem, recursive, list_merge))
    return sorted(d.values(), key=itemgetter(index))


def lists_mergeby(*terms, **kwargs):
    ''' Merge 2 or more lists by attribute 'index'. Optional parameters 'recursive' and 'list_merge'
        control the merging of the lists in values. The function merge_hash from ansible.utils.vars
        is used. To learn details on how to use the parameters 'recursive' and 'list_merge' see
        Ansible User's Guide chapter "Using filters to manipulate data" section "Combining
        hashes/dictionaries".

        Example:
        - debug:
            msg: "{{ list1|
                     community.general.lists_mergeby(list2,
                                                     'index',
                                                     recursive=True,
                                                     list_merge='append')|
                     list }}"
    '''

    recursive = kwargs.pop('recursive', False)
    list_merge = kwargs.pop('list_merge', 'replace')
    if kwargs:
        raise AnsibleFilterError("'recursive' and 'list_merge' are the only valid keyword arguments.")
    if len(terms) < 2:
        raise AnsibleFilterError("At least one list and index are needed.")

    # allow the user to do `[list1, list2, ...] | lists_mergeby('index')`
    flat_list = []
    for sublist in terms[:-1]:
        if not isinstance(sublist, Sequence):
            msg = ("All arguments before the argument index for community.general.lists_mergeby "
                   "must be lists. %s is %s")
            raise AnsibleFilterError(msg % (sublist, type(sublist)))
        if len(sublist) > 0:
            if all(isinstance(l, Sequence) for l in sublist):
                for item in sublist:
                    flat_list.append(item)
            else:
                flat_list.append(sublist)
    lists = flat_list

    if not lists:
        return []

    if len(lists) == 1:
        return lists[0]

    index = terms[-1]

    if not isinstance(index, string_types):
        msg = ("First argument after the lists for community.general.lists_mergeby must be string. "
               "%s is %s")
        raise AnsibleFilterError(msg % (index, type(index)))

    high_to_low_prio_list_iterator = reversed(lists)
    result = next(high_to_low_prio_list_iterator)
    for list in high_to_low_prio_list_iterator:
        result = list_mergeby(list, result, index, recursive, list_merge)

    return result


class FilterModule(object):
    ''' Ansible list filters '''

    def filters(self):
        return {
            'lists_mergeby': lists_mergeby,
        }



# -*- coding: utf-8 -*-
#
# Copyright: (c) 2019, Bojan Vitnik <bvitnik@mainstream.rs>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


def test_xapi_to_module_vm_power_state_bad_power_state(xenserver):
    """Tests that None is returned on bad power state."""
    assert xenserver.xapi_to_module_vm_power_state("bad") is None


def test_module_to_xapi_vm_power_state_bad_power_state(xenserver):
    """Tests that None is returned on bad power state."""
    assert xenserver.module_to_xapi_vm_power_state("bad") is None



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = '''
---
module: ohai
short_description: Returns inventory data from I(Ohai)
description:
     - Similar to the M(community.general.facter) module, this runs the I(Ohai) discovery program
       (U(https://docs.chef.io/ohai.html)) on the remote host and
       returns JSON inventory data.
       I(Ohai) data is a bit more verbose and nested than I(facter).
options: {}
notes: []
requirements: [ "ohai" ]
author:
    - "Ansible Core Team"
    - "Michael DeHaan (@mpdehaan)"
'''

EXAMPLES = '''
# Retrieve (ohai) data from all Web servers and store in one-file per host
ansible webservers -m ohai --tree=/tmp/ohaidata
'''
import json

from ansible.module_utils.basic import AnsibleModule


def main():
    module = AnsibleModule(
        argument_spec=dict()
    )
    cmd = ["/usr/bin/env", "ohai"]
    rc, out, err = module.run_command(cmd, check_rc=True)
    module.exit_json(**json.loads(out))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# (c) 2020, Alexei Znamensky <russoz@gmail.com>
# Copyright: (c) 2020, Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

import traceback
from functools import wraps

from ansible_collections.community.general.plugins.module_utils.mh.exceptions import ModuleHelperException


def cause_changes(on_success=None, on_failure=None):

    def deco(func):
        if on_success is None and on_failure is None:
            return func

        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                self = args[0]
                func(*args, **kwargs)
                if on_success is not None:
                    self.changed = on_success
            except Exception:
                if on_failure is not None:
                    self.changed = on_failure
                raise

        return wrapper

    return deco


def module_fails_on_exception(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            func(self, *args, **kwargs)
        except SystemExit:
            raise
        except ModuleHelperException as e:
            if e.update_output:
                self.update_output(e.update_output)
            self.module.fail_json(msg=e.msg, exception=traceback.format_exc(),
                                  output=self.output, vars=self.vars.output(), **self.output)
        except Exception as e:
            msg = "Module failed with exception: {0}".format(str(e).strip())
            self.module.fail_json(msg=msg, exception=traceback.format_exc(),
                                  output=self.output, vars=self.vars.output(), **self.output)
    return wrapper


def check_mode_skip(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.module.check_mode:
            return func(self, *args, **kwargs)
    return wrapper


def check_mode_skip_returns(callable=None, value=None):

    def deco(func):
        if callable is not None:
            @wraps(func)
            def wrapper_callable(self, *args, **kwargs):
                if self.module.check_mode:
                    return callable(self, *args, **kwargs)
                return func(self, *args, **kwargs)
            return wrapper_callable

        if value is not None:
            @wraps(func)
            def wrapper_value(self, *args, **kwargs):
                if self.module.check_mode:
                    return value
                return func(self, *args, **kwargs)
            return wrapper_value

    if callable is None and value is None:
        return check_mode_skip

    return deco
